from fastapi import APIRouter

from jina.helper import get_public_ip, get_internal_ip, get_full_version
from jina.logging.profile import used_memory_readable
from ...models import DaemonStatus
from ...stores import workspace_store, pea_store, pod_store, flow_store

router = APIRouter(tags=['daemon'])


@router.on_event('startup')
async def startup():
    """Start the process"""
    from ... import daemon_logger, jinad_args

    daemon_logger.info(
        f'''
Welcome to Jina daemon - the manager of distributed Jina
💬 Swagger UI     :\thttp://localhost:{jinad_args.port}/docs
📚 Redoc          :\thttp://localhost:{jinad_args.port}/redoc
🔒 Private address:\thttp://{get_internal_ip()}:{jinad_args.port}
🌐 Public address :\thttp://{get_public_ip()}:{jinad_args.port}'''
    )
    from jina import __ready_msg__

    daemon_logger.success(__ready_msg__)


@router.get(
    path='/',
)
async def _home():
    """
    The instruction HTML when user visits `/` directly


    .. #noqa: DAR201
    """
    return {}


@router.get(
    path='/status', summary='Get the status of the daemon', response_model=DaemonStatus
)
async def _status():
    _jina, _envs = get_full_version()
    return {
        'jina': _jina,
        'envs': _envs,
        'workspaces': workspace_store.status,
        'peas': pea_store.status,
        'pods': pod_store.status,
        'flows': flow_store.status,
        'used_memory': used_memory_readable(),
    }



from http import HTTPStatus
from typing import Union

import aiohttp

from .peas import AsyncPeaClient
from .mixin import AsyncToSyncMixin
from ..helper import if_alive, error_msg_from
from ..models.id import DaemonID, daemonize
from ..models.enums import UpdateOperation


class AsyncPodClient(AsyncPeaClient):
    """Async Client to create/update/delete Peods on remote JinaD"""

    _kind = 'pod'
    _endpoint = '/pods'

    @if_alive
    async def update(
        self, id: Union[str, 'DaemonID'], dump_path: str, *args, **kwargs
    ) -> str:
        """Update a Flow on remote JinaD (only rolling_update supported)

        :param id: flow id
        :param dump_path: path of dump from other flow
        :param args: positional args
        :param kwargs: keyword args
        :return: flow id
        """
        async with aiohttp.request(
            method='PUT',
            url=f'{self.store_api}/{daemonize(id, self._kind)}',
            params={
                'kind': UpdateOperation.ROLLING_UPDATE.value,
                'dump_path': dump_path,
            },
            timeout=self.timeout,
        ) as response:
            response_json = await response.json()
            if response.status != HTTPStatus.OK:
                error_msg = error_msg_from(response_json)
                self._logger.error(
                    f'{self._kind.title()} update failed as: {error_msg}'
                )
                return error_msg
            return response_json


class PodClient(AsyncToSyncMixin, AsyncPodClient):
    """Client to create/update/delete Pods on remote JinaD"""



import copy
from argparse import Namespace
from typing import Optional, Dict, Union, Set, List

import jina
from .k8slib import kubernetes_deployment, kubernetes_tools
from .. import BasePod
from ... import __default_executor__
from ...logging.logger import JinaLogger


class K8sPod(BasePod):
    """The K8sPod (KubernetesPod)  is used for deployments on Kubernetes."""

    def __init__(
        self, args: Union['Namespace', Dict], needs: Optional[Set[str]] = None
    ):
        super().__init__()
        self.args = args
        self.needs = needs or set()
        self.deployment_args = self._parse_args(args)

    def _parse_args(
        self, args: Namespace
    ) -> Dict[str, Optional[Union[List[Namespace], Namespace]]]:
        return self._parse_deployment_args(args)

    def _parse_deployment_args(self, args):
        parsed_args = {
            'head_deployment': None,
            'tail_deployment': None,
            'deployments': [],
        }
        parallel = getattr(args, 'parallel', 1)
        if parallel > 1:
            # reasons to separate head and tail from peas is that they
            # can be deducted based on the previous and next pods
            parsed_args['head_deployment'] = copy.copy(args)
            parsed_args['head_deployment'].uses = (
                args.uses_before or __default_executor__
            )
            parsed_args['tail_deployment'] = copy.copy(args)
            parsed_args['tail_deployment'].uses = (
                args.uses_after or __default_executor__
            )
            parsed_args['deployments'] = [args] * parallel
        else:
            parsed_args['deployments'] = [args]
        return parsed_args

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()
        self.join()

    @property
    def port_expose(self) -> int:
        """Not implemented"""
        raise NotImplementedError

    @property
    def host(self) -> str:
        """Not implemented"""
        raise NotImplementedError

    def _deploy_gateway(self, version):
        kubernetes_deployment.deploy_service(
            self.name,
            namespace=self.args.k8s_namespace,
            image_name=f'jinaai/jina:{version}-py38-standard',
            container_cmd='["jina"]',
            container_args=f'["gateway", '
            f'"--grpc-data-requests", '
            f'"--runtime-cls", "GRPCDataRuntime", '
            f'{kubernetes_deployment.get_cli_params(self.args, ("pod_role",))}]',
            logger=JinaLogger(f'deploy_{self.name}'),
            replicas=1,
            pull_policy='Always',
            init_container=None,
        )

    def _deploy_runtime(
        self, deployment_args, replicas, k8s_namespace, deployment_id, version
    ):
        image_name = kubernetes_deployment.get_image_name(deployment_args.uses)
        name_suffix = self.name + (
            '-' + str(deployment_id) if self.args.parallel > 1 else ''
        )
        dns_name = kubernetes_deployment.to_dns_name(name_suffix)
        init_container_args = kubernetes_deployment.get_init_container_args(self)
        uses_metas = kubernetes_deployment.dictionary_to_cli_param(
            {'pea_id': deployment_id}
        )
        uses_with = kubernetes_deployment.dictionary_to_cli_param(
            deployment_args.uses_with
        )
        uses_with_string = f'"--uses-with", "{uses_with}", ' if uses_with else ''
        if image_name == __default_executor__:
            image_name = f'jinaai/jina:{version}-py38-standard'
            container_args = (
                f'["pea", '
                f'"--uses", "BaseExecutor", '
                f'"--grpc-data-requests", '
                f'"--runtime-cls", "GRPCDataRuntime", '
                f'"--uses-metas", "{uses_metas}", '
                + uses_with_string
                + f'{kubernetes_deployment.get_cli_params(deployment_args)}]'
            )

        else:
            container_args = (
                f'["pea", '
                f'"--uses", "config.yml", '
                f'"--grpc-data-requests", '
                f'"--runtime-cls", "GRPCDataRuntime", '
                f'"--uses-metas", "{uses_metas}", '
                + uses_with_string
                + f'{kubernetes_deployment.get_cli_params(deployment_args)}]'
            )

        kubernetes_deployment.deploy_service(
            dns_name,
            namespace=k8s_namespace,
            image_name=image_name,
            container_cmd='["jina"]',
            container_args=container_args,
            logger=JinaLogger(f'deploy_{self.name}'),
            replicas=replicas,
            pull_policy='IfNotPresent',
            init_container=init_container_args,
            custom_resource_dir=getattr(self.args, 'k8s_custom_resource_dir', None),
        )

    def start(self) -> 'K8sPod':
        """Deploy the kubernetes pods via k8s Deployment and k8s Service.

        :return: self
        """
        kubernetes_tools.create(
            'namespace',
            {'name': self.args.k8s_namespace},
            custom_resource_dir=getattr(self.args, 'k8s_custom_resource_dir', None),
        )

        version = self._get_base_executor_version()
        if self.name == 'gateway':
            self._deploy_gateway(version)
        else:
            if self.deployment_args['head_deployment'] is not None:
                self._deploy_runtime(
                    self.deployment_args['head_deployment'],
                    1,
                    self.args.k8s_namespace,
                    'head',
                    version,
                )

            for i in range(self.args.parallel):
                deployment_args = self.deployment_args['deployments'][i]
                self._deploy_runtime(
                    deployment_args,
                    self.args.replicas,
                    self.args.k8s_namespace,
                    i,
                    version,
                )

            if self.deployment_args['tail_deployment'] is not None:
                self._deploy_runtime(
                    self.deployment_args['tail_deployment'],
                    1,
                    self.args.k8s_namespace,
                    'tail',
                    version,
                )
        return self

    def wait_start_success(self):
        """Not implemented. It should wait until the deployment is up and running"""
        pass

    def close(self):
        """Not implemented. It should delete the namespace of the flow"""
        pass

    def join(self):
        """Not implemented. It should wait to make sure deployments are properly killed."""
        pass

    @property
    def is_ready(self) -> bool:
        """Not implemented. It assumes it is ready.

        :return: True
        """
        return True

    @property
    def head_args(self) -> Namespace:
        """Head args of the pod.

        :return: namespace
        """
        return self.args

    @property
    def tail_args(self) -> Namespace:
        """Tail args of the pod

        :return: namespace
        """
        return self.args

    def is_singleton(self) -> bool:
        """The k8s pod is always a singleton

        :return: True
        """
        return True

    @property
    def num_peas(self) -> int:
        """Number of peas. Currently unused.

        :return: number of peas
        """
        return -1

    @property
    def head_zmq_identity(self) -> bytes:
        """zmq identity is not needed for k8s deployment

        :return: zmq identity
        """
        return b''

    @property
    def deployments(self) -> List[Dict]:
        """Deployment information which describes the interface of the pod.

        :return: list of dictionaries defining the attributes used by the routing table
        """
        res = []
        if self.args.name == 'gateway':
            name = kubernetes_deployment.to_dns_name(self.name)
            res.append(
                {
                    'name': f'{self.name}',
                    'head_host': f'{name}.{self.args.k8s_namespace}.svc.cluster.local',
                    'head_port_in': 8081,
                    'tail_port_out': 8082,
                    'head_zmq_identity': self.head_zmq_identity,
                }
            )
        else:
            if self.deployment_args['head_deployment'] is not None:
                name = kubernetes_deployment.to_dns_name(self.name + '_head')
                res.append(
                    {
                        'name': f'{self.name}_head',
                        'head_host': f'{name}.{self.args.k8s_namespace}.svc.cluster.local',
                        'head_port_in': 8081,
                        'tail_port_out': 8082,
                        'head_zmq_identity': self.head_zmq_identity,
                    }
                )
            for deployment_id, deployment_arg in enumerate(
                self.deployment_args['deployments']
            ):
                service_name = self.name + (
                    '-' + str(deployment_id)
                    if len(self.deployment_args['deployments']) > 1
                    else ''
                )
                name = kubernetes_deployment.to_dns_name(service_name)
                name_suffix = (
                    f'_{deployment_id}'
                    if len(self.deployment_args['deployments']) > 1
                    else ''
                )
                res.append(
                    {
                        'name': f'{self.name}{name_suffix}',
                        'head_host': f'{name}.{self.args.k8s_namespace}.svc.cluster.local',
                        'head_port_in': 8081,
                        'tail_port_out': 8082,
                        'head_zmq_identity': self.head_zmq_identity,
                    }
                )
            if self.deployment_args['tail_deployment'] is not None:
                name = kubernetes_deployment.to_dns_name(self.name + '_tail')
                res.append(
                    {
                        'name': f'{self.name}_tail',
                        'head_host': f'{name}.{self.args.k8s_namespace}.svc.cluster.local',
                        'head_port_in': 8081,
                        'tail_port_out': 8082,
                        'head_zmq_identity': self.head_zmq_identity,
                    }
                )
        return res

    def _get_base_executor_version(self):
        import requests

        url = 'https://registry.hub.docker.com/v1/repositories/jinaai/jina/tags'
        tags = requests.get(url).json()
        name_set = {tag['name'] for tag in tags}
        if jina.__version__ in name_set:
            return jina.__version__
        else:
            return 'master'



import re
import random
import operator
from collections import defaultdict
from typing import Dict, Optional, Union, Tuple, Any


if False:
    from .document import DocumentArray


class DocumentArraySearchOpsMixin:
    """ A mixin that provides search functionality to DocumentArrays"""

    _operators = {
        '<': operator.lt,
        '>': operator.gt,
        '==': operator.eq,
        '!=': operator.ne,
        '<=': operator.le,
        '>=': operator.ge,
    }

    def find(
        self,
        regexes: Dict[str, Union[str, re.Pattern]],
        traversal_paths: Tuple[str] = ('r',),
        operator: str = '>=',
        threshold: Optional[int] = None,
    ) -> 'DocumentArray':
        """
        Find Documents whose tag match the regular expressions in `regexes`.
        If `regexes` contain several regular expressions an `operator` can be used to
        specify a decision depending on the of regular expression matches specified by `value`.

        The supported operators are: ['<', '>', '==', '!=', '<=', '>=']

        Example: If `len(regexes)=3` then the documents from the DocumentArray will be accepted if
                 they match all 3 regular expressions.

        Example: If `len(regexes)=3`,  `value=2` and `operator='>='` then the documents
                 from the DocumentArray will be accepted if they match at least 2 regular expressions.

        :param regexes: Dictionary of the form {tag: Optional[str, regex]}
        :param traversal_paths: List specifying traversal paths
        :param operator: Operator used to accept/reject a document
        :param threshold: Number of regex that should match the operator to accept a Document.
                          If no value is provided `threshold=len(regexes)`.
        :return: DocumentArray with Documents that match the regexes
        """
        from .document import DocumentArray

        assert (
            operator in self._operators
        ), f'operator={operator} is not a valid operator from {self._operators.keys()}'

        operator_func = self._operators[operator]
        iterdocs = self.traverse_flat(traversal_paths)
        filtered = DocumentArray()

        threshold = threshold or len(regexes)

        for tag_name, regex in regexes.items():
            if isinstance(regex, str):
                regexes[tag_name] = re.compile(regex)

        for pos, doc in enumerate(iterdocs):
            counter = 0
            for tag_name, pattern in regexes.items():
                tag_value = doc.tags.get(tag_name, None)
                if tag_value:
                    if pattern.match(tag_value):
                        counter += 1
            if operator_func(counter, threshold):
                filtered.append(self[pos])

        return filtered

    def sample(self, k: int, seed: Optional[int] = None) -> 'DocumentArray':
        """random sample k elements from :class:`DocumentArray` without replacement.

        :param k: Number of elements to sample from the document array.
        :param seed: initialize the random number generator, by default is None. If set will
            save the state of the random function to produce certain outputs.
        :return: A sampled list of :class:`Document` represented as :class:`DocumentArray`.
        """

        if k > len(self):
            from ...helper import typename

            raise ValueError(
                f'Sample size can not be greater than the length of {typename(self)}, but {k} > {len(self)}'
            )
        if seed is not None:
            random.seed(seed)
        # NOTE, this could simplified to random.sample(self, k)
        # without getting indices and itemgetter etc.
        # however it's only work on DocumentArray, not DocumentArrayMemmap.
        indices = random.sample(range(len(self)), k)
        sampled = operator.itemgetter(*indices)(self)

        from .document import DocumentArray

        return DocumentArray(sampled)

    def shuffle(self, seed: Optional[int] = None) -> 'DocumentArray':
        """Randomly shuffle documents within the :class:`DocumentArray`.

        :param seed: initialize the random number generator, by default is None. If set will
            save the state of the random function to produce certain outputs.
        :return: The shuffled list of :class:`Document` represented as :class:`DocumentArray`.
        """
        from .document import DocumentArray

        return DocumentArray(self.sample(len(self), seed=seed))

    def split(self, tag: str) -> Dict[Any, 'DocumentArray']:
        """Split the `DocumentArray` into multiple DocumentArray according to the tag value of each `Document`.

        :param tag: the tag name to split stored in tags.
        :return: a dict where Documents with the same value on `tag` are grouped together, their orders
            are preserved from the original :class:`DocumentArray`.

        .. note::
            If the :attr:`tags` of :class:`Document` do not contains the specified :attr:`tag`,
            return an empty dict.
        """
        from .document import DocumentArray
        from ...helper import dunder_get

        rv = defaultdict(DocumentArray)
        for doc in self:
            if '__' in tag:
                value = dunder_get(doc.tags, tag)
            else:
                value = doc.tags.get(tag, None)

            if value is None:
                continue
            rv[value].append(doc)
        return dict(rv)



from typing import Dict
from http import HTTPStatus

import aiohttp

from .containers import ContainerStore
from ..excepts import PartialDaemon400Exception
from ..helper import if_alive, error_msg_from


class FlowStore(ContainerStore):
    """A Store of Flows spawned as Containers by Daemon"""

    _kind = 'flow'

    @if_alive
    async def _add(self, uri: str, params: Dict, **kwargs) -> Dict:
        """Sends `POST` request to `partial-daemon` to create a Flow.

        :param uri: uri of partial-daemon
        :param params: json payload to be sent
        :param kwargs: keyword args
        :raises PartialDaemon400Exception: if creation fails
        :return: response from partial-daemon
        """
        self._logger.debug(
            f'sending POST request to partial-daemon on {uri}/{self._kind}'
        )
        ports = kwargs.get('ports', [])
        async with aiohttp.request(
            method='POST',
            url=f'{uri}/{self._kind}',
            json={'flow': params, 'ports': ports},
        ) as response:
            response_json = await response.json()
            if response.status != HTTPStatus.CREATED:
                raise PartialDaemon400Exception(error_msg_from(response_json))
            return response_json

    @if_alive
    async def _delete(self, uri, **kwargs) -> Dict:
        """Sends a `DELETE` request to terminate the Flow & remove the container

        :param uri: uri of partial-daemon
        :param kwargs: keyword args
        :raises PartialDaemon400Exception: if deletion fails
        :return: response from partial-daemon
        """
        self._logger.debug(
            f'sending DELETE request to partial-daemon on {uri}/{self._kind}'
        )
        async with aiohttp.request(
            method='DELETE', url=f'{uri}/{self._kind}'
        ) as response:
            response_json = await response.json()
            if response.status != HTTPStatus.OK:
                raise PartialDaemon400Exception(error_msg_from(response_json))
            return response_json
